import datetime
import json
from pathlib import Path
from dataclasses import dataclass, asdict
from typing import Optional, List

from ..shared import *


@dataclass
class Author:
    """Name and email."""

    name: str
    email: str

    def dump(self) -> dict:
        return asdict(self)


@dataclass
class Dates:
    """Assignment dates."""

    assigned: datetime.datetime
    due: datetime.datetime

    def __init__(self, assigned: str, due: str):
        """Convert to datetime."""

        self.assigned = datetime.datetime.strptime(assigned, "%Y-%m-%d %H:%M:%S")
        self.due = datetime.datetime.strptime(due, "%Y-%m-%d %H:%M:%S")

    def dump(self) -> dict:
        return dict(
            assigned=self.assigned.strftime("%Y-%m-%d %H:%M:%S"),
            due=self.due.strftime("%Y-%m-%d %H:%M:%S"))


@dataclass
class Grading:
    """Grading description."""

    minutes: int
    automated: bool
    review: bool
    manual: bool

    def dump(self) -> dict:
        return asdict(self)


@dataclass
class Problem:
    """All problem data."""

    assignment: "Assignment"

    path: Path
    short: str
    number: int
    percentage: float
    directory: str  # Most specific directory containing relevant files

    title: str
    authors: List[Author]
    topics: List[str]
    grading: Grading
    notes: Optional[str] = None
    difficulty: Optional[str] = None

    @classmethod
    def load(cls, assignment: "Assignment", root: Path, reference: dict, number: int) -> "Problem":
        """Load a problem from the assignment path and reference."""

        path = root.joinpath(reference["path"])
        short = path.parts[-1]
        with path.joinpath(Files.PROBLEM).open() as file:
            data = json.load(file)

        authors = list(Author(**author) for author in data.pop("authors"))
        grading = Grading(**data.pop("grading"))
        percentage = reference["percentage"]
        directory = reference["directory"] if "directory" in reference else short

        return cls(
            assignment=assignment,
            path=path,
            short=short,
            number=number,
            percentage=percentage,
            authors=authors,
            grading=grading,
            directory=directory,
            **data)

    def dump(self) -> dict:
        return dict(
            short=self.short,
            percentage=self.percentage,
            title=self.title,
            authors=[author.dump() for author in self.authors],
            topics=self.topics,
            notes=self.notes,
            difficulty=self.difficulty,
            grading=self.grading.dump(),
            directory=self.directory)


@dataclass
class Assignment:
    """Contains assignment metadata."""

    path: Path
    short: str

    title: str
    authors: List[Author]
    dates: Dates
    problems: List[Problem]
    notes: Optional[str] = None

    @classmethod
    def load(cls, path: Path) -> "Assignment":
        """Load an assignment from a containing directory."""

        short = path.parts[-1]
        with path.joinpath(Files.ASSIGNMENT).open() as file:
            data = json.load(file)

        authors = list(Author(**author) for author in data.pop("authors"))
        dates = Dates(**data.pop("dates"))
        problem_references = data.pop("problems")

        self = cls(path, short, authors=authors, dates=dates, problems=[], **data)

        counter = 1
        for reference in problem_references:
            number = None
            if reference["percentage"] > 0:
                number = counter
                counter += 1
            self.problems.append(Problem.load(self, path, reference, number))

        return self

    def dump(self) -> dict:
        """Dump the assignment to JSON.

        This method is not symmetric to load, and should therefore
        probably not be called dump. It is intended for use in
        generating the grading schema.
        """

        return dict(
            short=self.short,
            title=self.title,
            authors=[author.dump() for author in self.authors],
            dates=self.dates.dump(),
            problems=[problem.dump() for problem in self.problems],
            notes=self.notes)



"""Grade is a combined library for grading C++ problem sets.

Grade is capable of testing three main aspects of C++ programs:
correctness, efficiency, and style. It includes tools for testing the
output of the code under various conditions, measuring its runtime
performance and memory usage, and running style checks.
"""

from pathlib import Path

ROOT = Path(__file__).absolute().parent
INCLUDE_PATH = ROOT.joinpath("include")



import importlib.util
import json
import sys
import timeit
from typing import Dict, Optional, Iterable, Iterator, Tuple
from pathlib import Path
from dataclasses import dataclass

from .grader import Grader
from .report import AssignmentReport, ProblemReport
from .resource import Context
from .exception import GraderException
from ..shared import *
from ..library.utility import timed
from ..log import log

PASSED = "\u2713"
FAILED = "\u2717"


def import_grader(tests_path: Path, grader_name: str = "grader") -> Grader:
    """Import a grader from a tests file."""

    sys.path.insert(0, str(tests_path.parent))
    spec = importlib.util.spec_from_file_location(f"{tests_path.parent}.tests", str(tests_path))
    tests = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(tests)
    grader = getattr(tests, grader_name)
    sys.path.pop(0)

    return grader


def print_reports(target_path: Path, reports: Dict[str, ProblemReport], options: dict):
    """Print reports to terminal."""

    print(f"{target_path}")
    for problem_short, report in reports.items():
        print(f"  {problem_short}: {report.statistics()}")
        if not options.get("concise"):
            for result in report.results:
                print(f"    {PASSED if result.complete and result.passed else FAILED} {result.task.name}")


@dataclass(eq=False)
class Manager:
    """A container for multiple graders in an assignment."""

    graders: Dict[str, Grader]
    schema: dict

    @classmethod
    def load(cls, grading_path: Path) -> Optional["Manager"]:
        """Load a manager from a grading artifact in the file system."""

        log.debug(f"loading manager from {grading_path}")
        log.debug("reading schema")
        with grading_path.joinpath(Files.GRADING).open() as file:
            schema = json.load(file)

        log.debug("assembling graders")
        graders = {}
        for problem_short in schema["problems"]:
            log.debug(f"importing grader for {problem_short}")
            grader = import_grader(grading_path.joinpath(problem_short, Files.TESTS))

            # Check dependencies
            try:
                grader.check()
            except GraderException:
                log.error(f"grader for {problem_short} failed to check")
                return None

            graders[problem_short] = grader

        return Manager(graders=graders, schema=schema)

    @timed(name="run", printer=log.info)
    def run_single(self, target_path: Path, **options) -> AssignmentReport:
        """Run all tests on a submission and return a dict of results."""

        log.info(f"running {target_path}")
        reports = AssignmentReport()

        for problem_short, grader in self.graders.items():
            log.debug(f"running problem {problem_short}")
            problem_schema = self.schema["problems"][problem_short]
            context = Context(
                target_path=target_path,
                problem_short=problem_short,
                problem_directory=target_path.joinpath(problem_schema["directory"]),
                options=options)

            try:
                reports[problem_short] = grader.run(context=context)
            except GraderException:
                return reports

        if options.get("report"):
            print_reports(target_path, reports, options)

        return reports

    def run_batch(self, target_paths: Iterable[Path], **options) -> Iterator[Tuple[Path, AssignmentReport]]:
        """Run multiple reports, map directory to report."""

        # Start timer
        start = timeit.default_timer()

        for target_path in target_paths:
            yield target_path, self.run_single(target_path, **options)

        elapsed = timeit.default_timer() - start
        log.info(f"finished batch in {round(elapsed, 5)} seconds")



import itertools
from typing import List, Dict, Iterable
from dataclasses import dataclass, field

from .report import ProblemReport
from .task import Task, Result
from .stage import GraderStage
from ..log import log

from .setup import SetupStage
from .test import TestStage
from .teardown import TeardownStage
from .configuration.sandbox import SandboxConfiguration
from .configuration.output import OutputConfiguration


def topological_sort_visit(task: Task, lookup: Dict[str, Task], marks: Dict[Task, int], result: List[Task]):
    """Visit a node."""

    if marks[task] == 2:
        return

    if marks[task] == 1:
        log.error("found cycle in task dependencies")
        raise ValueError()

    marks[task] = 1
    for dependency in task.dependencies:
        topological_sort_visit(lookup[dependency], lookup, marks, result)
    marks[task] = 2
    result.append(task)


def topological_sort(stage_tasks: Iterable[List[Task]]):
    """Order tasks by dependency."""

    lookup = {}
    marks = {}
    for tasks in stage_tasks:
        result = []
        for task in tasks:
            marks[task] = 0
            lookup[task.name] = task
        for task in tasks:
            if marks[task] != 2:
                topological_sort_visit(task, lookup, marks, result)
        tasks.clear()
        tasks.extend(result)


def collapse_tasks(stages: Iterable[GraderStage]) -> Iterable[Task]:
    """Wrapper for topological_sort."""

    return itertools.chain(*(stage.tasks for stage in stages))


def fulfills_dependencies(task: Task, report: ProblemReport):
    """Convenience."""

    return all(report.check(dependency) for dependency in task.dependencies)


def sanity_enabled_and_not_sanity(task: Task, resources: dict):
    """Convenience."""

    context = resources["context"]
    if "sanity" not in context.options or not context.options["sanity"]:
        return False
    if "sanity" in task.details and task.details["sanity"]:
        return False
    return True


@dataclass(eq=False)
class Grader:
    """A main class for grading runtime."""

    setup: SetupStage = field(default_factory=SetupStage)
    test: TestStage = field(default_factory=TestStage)
    teardown: TeardownStage = field(default_factory=TeardownStage)

    sandbox: SandboxConfiguration = field(default_factory=SandboxConfiguration)
    output: OutputConfiguration = field(default_factory=OutputConfiguration)

    @property
    def stages(self):
        return self.setup, self.test, self.teardown

    def check(self):
        """Topologically sort tasks, checking for cycles."""

        # Check dependencies
        log.debug("sorting grader tasks by dependency")
        topological_sort(stage.tasks for stage in self.stages)

        # Check task details
        for task in collapse_tasks(self.stages):
            self.output.check_task(task)

    def __run(self, tasks: List[Task], report: ProblemReport, resources: dict):
        """Execute sorted tasks, skipping if missing dependencies."""

        log.debug("running tasks")
        for task in tasks:
            log.debug(f"running task {task.name}")

            # Check conditions for whether this case is visible
            hidden = sanity_enabled_and_not_sanity(task, resources)

            # Run task if not hidden and dependencies are met
            if not hidden and fulfills_dependencies(task, report):
                try:
                    result = task.run(resources)
                except Result as r:
                    result = r

            # Otherwise take an incomplete result
            else:
                result = task.result_type.incomplete()

            result.task = task

            # Check result
            self.output.check_result(result)

            # Add to report
            report.add(result, hidden=hidden)

    def run(self, **resources) -> ProblemReport:
        """Build and test."""

        log.debug("setting up runtime")
        report = ProblemReport()
        resources.update(report=report, resources=resources)

        # Apply configuration
        log.debug("enabling configurators")
        self.sandbox.apply()

        # Run each stage
        for stage in self.stages:
            if len(stage.tasks) > 0:
                log.debug(f"starting stage {stage.name}")
                self.__run(stage.tasks, report, resources)
            else:
                log.debug(f"no tasks for stage {stage.name}")

        # Revert, trusting plugin
        log.debug("reverting configurators")
        self.sandbox.revert()

        return report

    def dump(self) -> dict:
        """Dump the tasks to something JSON serializable."""

        return {task.name: task.dump() for task in collapse_tasks(self.stages)}



from dataclasses import dataclass

from ...task import Result


@dataclass(init=False, eq=False)
class CheckResult(Result):
    """Result of a submission check."""

    kind = "check"

    def __init__(self, passed: bool, complete=True, **details):
        super().__init__(complete=complete, passed=passed, **details)



"""A package that provides tools for building assignments.

Curricula's build system takes an assignment composed of a set of a
set of problems, and separates the contents of each into combined
instructions, resources, grading, and solution artifacts. The
structure of an assignment is as follows:

assignment/
| README.md
| assignment.json

The structure of a problem is somewhat similar:

problem/
| README.md
| problem.json
| assets/
| | ...
| grading/
| | README.md
| | tests.py
| | ...
| resources/
| | ...
| solution/
| | README.md
| | ...

Typically, problems are located in a sub-folder assignment/problem/.
However, the problems are linked to assignments in the assignment.json
file via relative path, so the problem directory could be anywhere.
"""

import jinja2
import json
from pathlib import Path, PurePath
from typing import Dict, Union, List, Callable
from dataclasses import dataclass

from .models import Assignment, Problem
from ..shared import Files, Paths
from ..library.template import jinja2_create_environment
from ..library import files
from ..library.utility import timed
from ..log import log

from ..grade.manager import import_grader


@dataclass(repr=False, eq=False)
class Context:
    """Build context.

    The build context is simply a container for information about the
    build that's passed to each step.
    """

    environment: jinja2.Environment
    assignment_path: Path
    artifacts_path: Path
    options: Dict[str, str]


def compile_readme(
        context: Context,
        assignment: Assignment,
        template_relative_path: str,
        destination_path: Path) -> Path:
    """Compile a README from an assignment.

    This function returns the final path of the README, which may be
    different if the provided destination is a directory.
    """

    log.debug(f"compiling readme for {destination_path}")
    template = context.environment.get_template(f"template:build/{template_relative_path}")
    if destination_path.is_dir():
        destination_path = destination_path.joinpath(Files.README)
    with destination_path.open("w") as file:
        file.write(template.render(assignment=assignment))
    return destination_path


def merge_contents(
        assignment: Assignment,
        contents_relative_path: Path,
        destination_path: Path,
        filter_problems: Callable[[Problem], bool] = None):
    """Compile subdirectories from problems into a single directory."""

    log.debug(f"merging contents to {destination_path}")

    destination_path.mkdir(exist_ok=True)

    # First copy any assignment-wide resources
    assignment_contents_path = assignment.path.joinpath(contents_relative_path)
    if assignment_contents_path.exists():
        files.copy_directory(assignment_contents_path, destination_path)

    # Overwrite with problem contents, enable filtration
    for problem in filter(filter_problems, assignment.problems):
        problem_contents_path = problem.path.joinpath(contents_relative_path)
        if problem_contents_path.exists():
            files.copy_directory(problem_contents_path, destination_path, merge=True)


def aggregate_contents(
        assignment: Assignment,
        contents_relative_path: Path,
        destination_path: Path,
        filter_problems: Callable[[Problem], bool] = None,
        rename: Callable[[Problem], str] = None) -> List[Path]:
    """Compile subdirectories from problems to respective directories.

    Different from merge in that the result is a directory of
    directories. Returns a list of the resultant folders that were
    copied into the destination.
    """

    log.debug(f"aggregating contents to {destination_path}")

    destination_path.mkdir(exist_ok=True)

    # First compile assignment-wide assets
    assignment_contents_path = assignment.path.joinpath(contents_relative_path)
    if assignment_contents_path.exists():
        files.copy_directory(assignment_contents_path, destination_path)

    # Copy per problem, enable filtration
    copied_paths = []
    for problem in filter(filter_problems, assignment.problems):
        problem_contents_path = problem.path.joinpath(contents_relative_path)
        if problem_contents_path.exists():
            problem_destination_path = destination_path.joinpath(problem.short if rename is None else rename(problem))
            copied_paths.append(problem_destination_path)
            files.copy_directory(problem_contents_path, problem_destination_path)

    return copied_paths


def build_instructions(context: Context, assignment: Assignment):
    """Compile instruction readme and assets."""

    log.debug("compiling instructions")
    instructions_path = context.artifacts_path.joinpath(Paths.INSTRUCTIONS)
    instructions_path.mkdir(exist_ok=True)
    compile_readme(context, assignment, "instructions/assignment.md", instructions_path)
    merge_contents(assignment, Paths.ASSETS, instructions_path.joinpath(Paths.ASSETS))


def build_resources(context: Context, assignment: Assignment):
    """Compile resources files."""

    log.debug("compiling resources")
    resources_path = context.artifacts_path.joinpath(Paths.RESOURCES)
    resources_path.mkdir(exist_ok=True)
    aggregate_contents(assignment, Paths.RESOURCES, resources_path, rename=lambda p: p.directory)


def build_solution_readme(context: Context, assignment: Assignment, path: Path):
    """Generate the composite cheat sheet README."""

    log.debug("building cheat sheet")
    assignment_template = context.environment.get_template("template:build/solution/assignment.md")
    with path.joinpath(Files.README).open("w") as file:
        file.write(assignment_template.render(assignment=assignment))


def build_solution_code(assignment: Assignment, path: Path):
    """Compile only submission files of the solution."""

    log.debug("assembling solution code")
    copied_paths = aggregate_contents(assignment, Paths.SOLUTION, path, rename=lambda p: p.directory)

    # Delete extra READMEs
    for copied_path in copied_paths:
        readme_path = copied_path.joinpath(Files.README)
        if readme_path.exists():
            files.delete(readme_path)


def build_solution(context: Context, assignment: Assignment):
    """Compile cheatsheets."""

    log.debug("compiling solution")
    solution_path = context.artifacts_path.joinpath(Paths.SOLUTION)
    solution_path.mkdir(exist_ok=True)
    build_solution_readme(context, assignment, solution_path)
    build_solution_code(assignment, solution_path)


def build_grading_readme(context: Context, assignment: Assignment, path: Path):
    """Aggregate README for rubric."""

    log.debug("building grading instructions")
    assignment_template = context.environment.get_template("template:build/grading/assignment.md")
    with path.joinpath(Files.README).open("w") as file:
        file.write(assignment_template.render(assignment=assignment))


def generate_grading_schema(grading_path: Path, assignment: Assignment) -> dict:
    """Generate a JSON schema describing the grading package.

    This method requires the grading artifact to already have been
    aggregated, as it has to access the individual problem graders to
    dump their task summaries.
    """

    schema = dict(title=assignment.title, short=assignment.short, problems=dict())
    for problem in assignment.problems:
        if problem.grading.automated:
            grader = import_grader(grading_path.joinpath(problem.short, Files.TESTS))
            schema["problems"][problem.short] = dict(
                title=problem.title,
                directory=problem.directory,
                percentage=problem.percentage,
                tasks=grader.dump())
    return schema


def build_grading_schema(assignment: Assignment, path: Path):
    """Generate a JSON data file with grading metadata."""

    log.debug("building grading schema")
    with path.joinpath(Files.GRADING).open("w") as file:
        json.dump(generate_grading_schema(path, assignment), file, indent=2)


def build_grading(context: Context, assignment: Assignment):
    """Compile rubrics."""

    log.debug("compiling grading")

    grading_path = context.artifacts_path.joinpath(Paths.GRADING)
    grading_path.mkdir(exist_ok=True)
    build_grading_readme(context, assignment, grading_path)
    copied_paths = aggregate_contents(
        assignment,
        Paths.GRADING,
        grading_path,
        filter_problems=lambda p: p.grading.automated)

    # Delete extra READMEs
    for copied_path in copied_paths:
        readme_path = copied_path.joinpath(Files.README)
        if readme_path.exists():
            files.delete(readme_path)

    build_grading_schema(assignment, grading_path)


def build_index(context: Context, assignment: Assignment):
    """Dump the assignment into the artifacts."""

    log.debug("writing index to artifacts")
    with context.artifacts_path.joinpath("index.json").open("w") as file:
        json.dump(assignment.dump(), file, indent=2)


@jinja2.environmentfilter
def get_readme(environment: jinja2.Environment, item: Union[Problem, Assignment], *component: str) -> str:
    """Render a README with options for nested path."""

    readme_path = "/".join(component + (Files.README,))

    try:
        if isinstance(item, Assignment):
            return environment.get_template(f"assignment:{readme_path}").render(assignment=item)
        elif isinstance(item, Problem):
            return environment.get_template(f"problem/{item.short}:{readme_path}").render(
                assignment=item.assignment, problem=item)
    except jinja2.exceptions.TemplateNotFound as exception:
        log.error(f"error finding {exception}")
        return ""


def has_readme(item: Union[Problem, Assignment], *component: str) -> bool:
    """Check whether a problem has a solution README."""

    return item.path.joinpath(*component, Files.README).exists()


BUILD_STEPS = (
    build_instructions,
    build_resources,
    build_solution,
    build_grading,
    build_index)


@timed("build", printer=log.info)
def build(template_path: Path, assignment_path: Path, artifacts_path: Path, **options):
    """Build the assignment at a given path."""

    log.info(f"building {assignment_path} to {artifacts_path}")

    # Load the assignment object
    log.debug("loading assignment")
    assignment = Assignment.load(assignment_path)

    # Set up templating
    problem_template_paths = {f"problem/{problem.short}": problem.path for problem in assignment.problems}
    environment = jinja2_create_environment(
        template=template_path,
        assignment=assignment_path,
        **problem_template_paths)
    environment.filters.update(get_readme=get_readme, has_readme=has_readme)

    # Define context
    log.debug("setting context")
    context = Context(environment, assignment_path, artifacts_path, options)
    environment.globals["context"] = context

    # Create output directory
    artifacts_path.mkdir(exist_ok=True, parents=True)

    # Build
    for step in BUILD_STEPS:
        step(context, assignment)



from typing import List, Dict
from dataclasses import dataclass, field

from .resource import Resource
from .task import Result


@dataclass(eq=False)
class ProblemReportStatistics:
    """Rudimentary sums from the report results."""

    tasks_total: int = 0
    tasks_complete: int = 0
    tasks_passed: int = 0
    tests_total: int = 0
    tests_complete: int = 0
    tests_passed: int = 0

    def __str__(self) -> str:
        """Format nicely."""

        tasks_percentage = round(self.tasks_complete/(self.tasks_complete or 1) * 100, 2)
        tests_percentage = round(self.tests_passed/(self.tests_complete or 1) * 100, 2)
        return (f"{self.tasks_complete}/{self.tasks_complete} tasks complete ({tasks_percentage}%), "
                f"{self.tests_passed}/{self.tests_total} tests passed ({tests_percentage}%)")


@dataclass(eq=False)
class ProblemReport(Resource):
    """The final report returned by the testing framework."""

    lookup: Dict[str, Result] = field(default_factory=dict)
    results: List[Result] = field(default_factory=list)

    def check(self, task_name: str) -> bool:
        """Check the result of a task."""

        result = self.lookup[task_name]
        return result.complete and result.passed

    def add(self, result: Result, hidden: bool = False):
        """Add a result to the report."""

        self.lookup[result.task.name] = result
        if not hidden:
            self.results.append(result)

    def dump(self) -> dict:
        """Dump the result to a serializable format."""

        return {result.task.name: result.dump() for result in self.results}

    def statistics(self) -> ProblemReportStatistics:
        """Run the numbers."""

        statistics = ProblemReportStatistics()
        for result in self.results:
            statistics.tasks_total += 1
            if result.complete:
                statistics.tasks_complete += 1
            if result.passed:
                statistics.tasks_passed += 1
            if result.task.stage == "test":
                statistics.tests_total += 1
                if result.complete:
                    statistics.tests_complete += 1
                if result.passed:
                    statistics.tests_passed += 1
        return statistics


class AssignmentReport(dict):
    """Aggregation of problem reports."""

    def dump(self) -> dict:
        return {problem_short: problem_report.dump() for problem_short, problem_report in self.items()}



import subprocess
import timeit
import os
import time
import signal

try:
    import fcntl
except ImportError:
    fcntl = None

from typing import Optional, Tuple, Callable, IO
from dataclasses import dataclass, asdict, field
from contextlib import contextmanager
from pathlib import Path


@dataclass(eq=False)
class RuntimeException:
    """Error that occurs during process runtime."""

    description: str
    error_number: Optional[int] = None

    @classmethod
    def from_os_error(cls, error: OSError) -> "RuntimeException":
        """Create from OS error."""

        return RuntimeException(
            description="executable format error" if error.errno == 8 else "failed to run executable",
            error_number=error.errno)

    def dump(self) -> dict:
        """Serialize."""

        return asdict(self)


@dataclass(eq=False)
class Interaction:
    """Container for single interaction with a process."""

    args: Tuple[str, ...]
    cwd: Optional[Path]

    elapsed: Optional[float] = None
    stdin: Optional[bytes] = None
    stdout: Optional[bytes] = None
    stderr: Optional[bytes] = None

    def dump(self) -> dict:
        """Make the runtime JSON serializable."""

        dump = asdict(self)
        for field_name in "stdout", "stdin", "stderr":
            if dump[field_name] is not None:
                dump[field_name] = dump[field_name].decode(errors="replace")
        if dump["cwd"]:
            dump["cwd"] = str(dump["cwd"])
        return dump


@dataclass(eq=False)
class Runtime(Interaction):
    """Runtime data extracted from running an external process."""

    code: Optional[int] = None

    timeout: Optional[float] = None
    timed_out: bool = False

    raised_exception: bool = False
    exception: Optional[RuntimeException] = None


class InteractiveStreamTimeoutExpired(RuntimeError):
    buffer: bytes

    def __init__(self, buffer: bytes):
        self.buffer = buffer


class InteractiveStream:
    """Custom IO stream for Interactive."""

    file: IO[bytes]
    history: bytes

    POLL: float = 0.001

    def __init__(self, file: IO[bytes], read: bool, write: bool):
        """Create the stream."""

        self.file = file
        self.history = b""

        if read:
            self._set_nonblock()

        # Disable methods
        if not read:
            self.read = None
        if not write:
            self.write = None

    def _set_nonblock(self):
        """Disable blocking for read."""

        flags = fcntl.fcntl(self.file, fcntl.F_GETFL)
        fcntl.fcntl(self.file, fcntl.F_SETFL, flags | os.O_NONBLOCK)

    def _read_nonblock(self) -> Optional[bytes]:
        """Read or return None if no data."""

        data = self.file.read()
        if data is not None:
            self.history += data
        return data

    def _read_block(self, condition: Callable[[bytes], bool] = None, timeout: float = None) -> Optional[bytes]:
        """Block until something besides None is returned."""

        buffer = b""

        timeout_time = None
        if timeout is not None:
            timeout_time = timeit.default_timer() + timeout

        while True:
            data = self.file.read()
            if data is not None:
                buffer += data
                if condition is None or condition(buffer):
                    break
            if timeout is not None and timeit.default_timer() >= timeout_time:
                self.history += buffer
                raise InteractiveStreamTimeoutExpired(buffer=buffer)
            time.sleep(self.POLL)

        self.history += buffer
        return buffer

    def read(
            self,
            block: bool = True,
            condition: Callable[[bytes], bool] = None,
            timeout: float = None) -> Optional[bytes]:
        """Read from a stream.

        If blocking is disabled, may return None. If condition is not
        None, block until condition is satisfied. If timeout is not
        None, break after timeout and return buffer or None.
        """

        if block:
            return self._read_block(condition=condition, timeout=timeout)
        return self._read_nonblock()

    def write(
            self,
            *values: bytes,
            sep: bytes = b" ",
            end: bytes = b"\n",
            flush: bool = True):
        """Write to the stream like traditional print."""

        data = sep.join(values) + end
        self.file.write(data)
        self.history += data
        if flush:
            try:
                self.file.flush()
            except BrokenPipeError:
                pass


@dataclass(eq=False)
class Interactive:
    """An interactive runtime session."""

    _args: Tuple[str, ...]
    _process: subprocess.Popen
    _start_time: float
    cwd: Optional[Path]
    stdin: InteractiveStream
    stdout: InteractiveStream
    stderr: InteractiveStream

    _recording: Optional[Interaction] = None

    def __init__(self, args: Tuple[str, ...], cwd: Path = None):
        """Start up the new process."""

        self._args = args
        self._process = subprocess.Popen(
            args=args,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            stdin=subprocess.PIPE,
            preexec_fn=config.process_setup,
            cwd=str(cwd) if cwd is not None else None)
        self.cwd = cwd
        self.stdin = InteractiveStream(self._process.stdin, read=False, write=True)
        self.stdout = InteractiveStream(self._process.stdout, read=True, write=False)
        self.stderr = InteractiveStream(self._process.stdout, read=True, write=False)
        self._start_time = timeit.default_timer()

    def poll(self) -> bool:
        """Check whether the interactive has terminated."""

        return self._process.poll() is None

    @contextmanager
    def recording(self) -> Interaction:
        """Record a frame of all process streams."""

        if self._recording is not None:
            raise RuntimeError("Cannot make multiple runtime recordings at once!")

        partial = Interaction(args=self._args, cwd=self.cwd)
        stdin_index = len(self.stdin.history)
        stdout_index = len(self.stdout.history)
        stderr_index = len(self.stderr.history)
        start_time = timeit.default_timer()

        yield partial

        # Collect everything that changed
        partial.elapsed = timeit.default_timer() - start_time
        partial.stdin = self.stdin.history[stdin_index:]
        partial.stdout = self.stdout.history[stdout_index:]
        partial.stderr = self.stderr.history[stderr_index:]

    def close(self, timeout: float = None) -> Runtime:
        """Block until exit."""

        raised_exception = False
        exception = None
        timed_out = False
        stdout = b""
        stderr = b""

        try:
            stdout, stderr = self._process.communicate(timeout=timeout)
        except subprocess.TimeoutExpired:
            timed_out = True
        except OSError as error:
            raised_exception = True
            exception = RuntimeException.from_os_error(error)

        stop_time = timeit.default_timer()
        return Runtime(
            args=self._args,
            cwd=self.cwd,
            timeout=timeout,
            code=self._process.returncode,
            elapsed=stop_time - self._start_time,
            stdin=self.stdin.history,
            stdout=self.stdout.history + stdout,
            stderr=self.stderr.history + stderr,
            raised_exception=raised_exception,
            exception=exception,
            timed_out=timed_out)


def demote_user(user_uid: int, user_gid: int):
    """Set the user of the process."""

    os.setgid(user_gid)
    os.setuid(user_uid)


@dataclass(eq=False)
class SubprocessConfiguration:
    """Global settings."""

    custom_process_setup: Optional[Callable[[], None]] = None

    demote_user_enabled: bool = False
    demote_user_uid: int = field(init=False)
    demote_user_gid: int = field(init=False)

    def set_custom_process_setup(self, custom_process_setup: Callable[[], None]):
        """Can be used as a decorator."""

        self.custom_process_setup = custom_process_setup

    def clear_custom_process_setup(self):
        """Return to default."""

        self.custom_process_setup = None

    def enable_demote_user(self, uid: int, gid: int):
        """Globally enable user demotion in run."""

        self.demote_user_enabled = True
        self.demote_user_uid = uid
        self.demote_user_gid = gid

    def disable_demote_user(self):
        """Globally disable user demotion."""

        self.demote_user_enabled = False

    def process_setup(self):
        """Do not overwrite."""

        if self.custom_process_setup is not None:
            self.custom_process_setup()
        if self.demote_user_enabled:
            demote_user(self.demote_user_uid, self.demote_user_gid)


config = SubprocessConfiguration()


def run(*args: str, stdin: bytes = None, timeout: float = None, cwd: Path = None) -> Runtime:
    """Run an executable with a list of command line arguments.

    The provided path must be absolute in order to properly execute
    the program. Args provided are passed as they would be from the
    command line. The timeout is measured in seconds.

    The process_setup callable is invoked within the spawned process
    prior to the execution of the command.
    """

    # Spawn the process, access stdout and stderr
    try:
        if stdin is not None:
            process = subprocess.Popen(
                args,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                stdin=subprocess.PIPE,
                preexec_fn=config.process_setup,
                cwd=str(cwd) if cwd is not None else None)
        else:
            process = subprocess.Popen(
                args,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                preexec_fn=config.process_setup,
                cwd=str(cwd) if cwd is not None else None)

    # Catch common errors
    except OSError as error:
        exception = RuntimeException.from_os_error(error)
        return Runtime(args=args, cwd=cwd, timeout=timeout, stdin=stdin, raised_exception=True, exception=exception)
    except ValueError:
        exception = RuntimeException(description="failed to open process")
        return Runtime(args=args, cwd=cwd, timeout=timeout, stdin=stdin, raised_exception=True, exception=exception)
    except subprocess.SubprocessError as exception:
        exception = RuntimeException(description=str(exception))
        return Runtime(args=args, cwd=cwd, timeout=timeout, stdin=stdin, raised_exception=True, exception=exception)

    # Wait for the process to finish with timeout
    start = timeit.default_timer()
    try:
        stdout, stderr = process.communicate(input=stdin, timeout=timeout)
    except subprocess.TimeoutExpired:
        process.kill()

        # Recover data
        try:
            stdout, stderr = process.communicate(timeout=1)
        except subprocess.TimeoutExpired:
            stdout, stderr = None, None

        return Runtime(args=args, cwd=cwd, timeout=timeout, stdin=stdin, stdout=stdout, stderr=stderr, timed_out=True)

    # Check elapsed
    elapsed = timeit.default_timer() - start
    return Runtime(
        args=args,
        cwd=cwd,
        timeout=timeout,
        code=process.returncode,
        elapsed=elapsed,
        stdin=stdin,
        stdout=stdout,
        stderr=stderr)


def interactive(*args: str) -> Interactive:
    return Interactive(args=args)



from ...task import Result


class ComplexityResult(Result):
    """The result of a correctness case."""

    kind = "correctness"

    def __init__(self, passed: bool, complete: bool = True, **details):
        super().__init__(complete=complete, passed=passed, details=details)



import jinja2
import logging
from decimal import Decimal
from pathlib import Path
from typing import Any

root = Path(__file__).absolute().parent
log = logging.getLogger("curricula")


def percentage(d: Any, digits: int = 1) -> str:
    """Convert a float to a nice-looking percentage."""

    converted = d if isinstance(d, Decimal) else Decimal(str(d))
    converted *= 100
    if converted == converted.to_integral_value():
        return f"{int(converted)}%"
    return f"{round(converted, digits)}%"


JINJA2_FILTERS = {
    "datetime": lambda d: d.strftime("%B %d, %Y at %H:%M"),
    "date": lambda d: d.strftime("%B %d, %Y"),
    "percentage": percentage,
}


def jinja2_create_environment(**template_paths: Path) -> jinja2.Environment:
    """Configure a jinja2 environment."""

    log.debug("creating jinja2 environment")

    # Create a loader in the order of arguments
    loader = jinja2.PrefixLoader({
        key: jinja2.FileSystemLoader(str(path)) for key, path in template_paths.items()
    }, delimiter=":")

    # Actually create the environment, we use square brackets to avoid
    # clashing with Jekyll
    environment = jinja2.Environment(
        loader=loader,
        block_start_string="[%",
        block_end_string="%]",
        variable_start_string="[[",
        variable_end_string="]]",
        comment_start_string="[#",
        comment_end_string="#]",
        autoescape=False,
        keep_trailing_newline=True)

    # Custom filters
    environment.filters.update(JINJA2_FILTERS)

    return environment



import json
import jsonschema
from pathlib import Path

from .models import Assignment
from ..shared import *
from ..log import log

root = Path(__file__).absolute().parent

with root.joinpath("schema", "assignment.schema.json").open() as _file:
    ASSIGNMENT_SCHEMA = json.load(_file)
with root.joinpath("schema", "problem.schema.json").open() as _file:
    PROBLEM_SCHEMA = json.load(_file)


def validate_problem_directory(problem_path: Path):
    """Validate with jsonschema."""

    try:
        jsonschema.validate(problem_path.joinpath(Files.PROBLEM).read_text(), PROBLEM_SCHEMA)
    except jsonschema.ValidationError as exception:
        log.error(f"invalid schema in problem {problem_path}: {exception}")
        raise


def validate_assignment_directory(assignment_path: Path):
    """Validate with jsonschema."""

    try:
        jsonschema.validate(assignment_path.joinpath(Files.ASSIGNMENT).read_text(), PROBLEM_SCHEMA)
    except jsonschema.ValidationError as exception:
        log.error(f"invalid schema in assignment {assignment_path}: {exception}")
        raise


def validate_assignment(assignment_path: Path):
    """Validators per assignment."""

    validate_assignment_directory(assignment_path)

    assignment = Assignment.load(assignment_path)
    for problem in assignment.problems:
        validate_problem_directory(problem.path)


def validate(assignment_path: Path):
    """Do a bunch of checks """

    validate_assignment(assignment_path)



from ..stage import GraderStage
# from .code import CodeResult
from .complexity import ComplexityResult
from .correctness import CorrectnessResult
from .memory import MemoryResult


class TestStage(GraderStage):
    """Test endpoints."""

    name = "test"

    # def code(self, **details):
    #     """Code quality."""
    #
    #     return self.create_registrar(details, CodeResult)

    def complexity(self, **details):
        """Runtime complexity."""

        return self.create_registrar(details, ComplexityResult)

    def correctness(self, **details):
        """Program correctness."""

        return self.create_registrar(details, CorrectnessResult)

    def memory(self, **details):
        """Resource allocation."""

        return self.create_registrar(details, MemoryResult)



from dataclasses import dataclass

from ...task import Result


@dataclass(init=False, eq=False)
class BuildResult(Result):
    """Returned from a build task."""

    kind = "build"

    def __init__(self, passed: bool, complete: bool = True, **details):
        super().__init__(complete=complete, passed=passed, details=details)



from typing import List, Type, Callable

from .task import Task, Runnable, Result
from ..log import log

__all__ = ("GraderStage", "Registrar")

Registrar = Callable[[Runnable], Runnable]
RegistrarFactory = Callable[[dict, Type[Result]], Registrar]


def create_registrar_factory(stage: str, task_list: List[Task]) -> RegistrarFactory:
    """A third-level decorator to reuse code."""

    def create_registrar(details: dict, result_type: Type[Result]) -> Registrar:
        """Differentiate based on the kind of result."""

        def decorator(runnable: Runnable) -> Runnable:
            """Put the function in a correctness object."""

            name = details.pop("name", runnable.__qualname__)
            description = details.pop("description", runnable.__doc__)

            # Get dependencies
            if "dependencies" in details:
                if "dependency" in details:
                    log.error("only one of dependency and dependencies may be specified")
                    raise ValueError()
                dependencies = details.pop("dependencies")
            elif "dependency" in details:
                dependencies = (details.pop("dependency"),)
            else:
                dependencies = ()

            for existing_task in task_list:
                if existing_task.name == name:
                    raise ValueError(f"Duplicate task name \"{name}\"")

            # Create task, append
            task_list.append(Task(
                name=name,
                description=description,
                stage=stage,
                kind=result_type.kind,
                dependencies=dependencies,
                runnable=runnable,
                details=details,
                result_type=result_type))
            return runnable

        return decorator

    return create_registrar


class GraderStage:
    """Management for setup, test, and teardown stages."""

    name: str
    tasks: List[Task]
    create_registrar: RegistrarFactory

    def __init__(self):
        self.tasks = []
        self.create_registrar = create_registrar_factory(stage=self.name, task_list=self.tasks)



from . import GraderConfiguration
from ...library import process
from ...log import log

try:
    import pwd
except ImportError:
    pwd = None
    log.warning("pwd module is unavailable on this platform")


class SandboxConfiguration(GraderConfiguration):
    """Sandbox tools for the grader."""

    user_demotion_enabled: bool
    user_demotion_username: str

    def __init__(self):
        """Initialize the sandbox."""

        self.user_demotion_enabled = False

    def enable_user_demotion(self, *, username: str = "grader"):
        """Add a setup step that demotes the process user."""

        self.user_demotion_enabled = True
        self.user_demotion_username = username

    def apply(self):
        """Enable configuration."""

        if self.user_demotion_enabled:
            log.debug("enabling user demotion")

            try:
                record = pwd.getpwnam(self.user_demotion_username)
            except KeyError:
                log.error("grader system account not available")
                return

            process.config.enable_demote_user(record.pw_uid, record.pw_gid)

    def revert(self):
        """Revert changes."""

        if self.user_demotion_enabled:
            log.debug("reverting user demotion")
            process.config.disable_demote_user()



import json
from typing import Dict, List, Optional, Iterable
from pathlib import Path
from dataclasses import dataclass, field
from decimal import Decimal

from ...library.template import jinja2_create_environment
from ...shared import Files


def sum_weights(tasks: Iterable[dict]) -> float:
    return sum(Decimal(str(task["details"].get("weight", "1"))) for task in tasks)


@dataclass(eq=False)
class ProblemSummary:
    """A summary of the results from a problem's cases."""

    # Main tests
    tests_total: int = 0
    tests_correct: List[dict] = field(default_factory=list)
    tests_incorrect: List[dict] = field(default_factory=list)

    # Setup problems
    setup_failed: bool = False
    setup_failed_task: Optional[str] = None
    setup_error: Optional[str] = None

    @property
    def tests_fraction(self) -> str:
        """Format a fraction."""

        numerator = sum_weights(self.tests_correct)
        denominator = numerator + sum_weights(self.tests_incorrect)
        return f"{numerator}/{denominator}"

    @property
    def tests_percentage(self) -> Decimal:
        """Compute the percentage."""

        if self.tests_total == 0:
            return Decimal("0")
        numerator = sum_weights(self.tests_correct)
        denominator = numerator + sum_weights(self.tests_incorrect)
        return Decimal(numerator) / Decimal(denominator)


@dataclass(eq=False)
class ReportSummary:
    """Summary of problems."""

    problems: Dict[str, ProblemSummary] = field(default_factory=dict)


def summarize(grading_schema: dict, report: dict) -> ReportSummary:
    """Do a summary for a single student."""

    summary = ReportSummary()
    for problem_short, problem in grading_schema["problems"].items():
        problem_summary = summary.problems[problem_short] = ProblemSummary()
        for task_name, task in problem["tasks"].items():
            result = report[problem_short][task_name]

            # Diagnose any issues in setup
            if task["stage"] == "setup":
                if not result["complete"] or not result["passed"]:
                    problem_summary.setup_failed = True
                    problem_summary.setup_failed_task = task_name
                    if "error" in result["details"]:
                        problem_summary.setup_error = result["details"]["error"]
                    elif "runtime" in result["details"]:
                        problem_summary.setup_error = result["details"]["runtime"]["stderr"]

            # Problem summary
            elif task["stage"] == "test":
                problem_summary.tests_total += 1
                if result["complete"] and result["passed"]:
                    problem_summary.tests_correct.append(task)
                else:
                    problem_summary.tests_incorrect.append(task)

    return summary


def format_report_markdown(grading_path: Path, template_path: Path, report_path: Path) -> str:
    """Return a formatted markdown report."""

    environment = jinja2_create_environment()
    with grading_path.joinpath(Files.GRADING).open() as file:
        grading_schema = json.load(file)
    with report_path.open() as file:
        report = json.load(file)
    summary = summarize(grading_schema, report)

    environment.globals.update(schema=grading_schema, summary=summary)
    report_template = environment.from_string(template_path.read_text())
    return report_template.render() + "\n"



from dataclasses import dataclass, field
from typing import Dict, Any

from . import GraderConfiguration
from ..exception import GraderException
from ..task import Task, Result
from ...log import log

Nothing = object()


@dataclass(eq=False)
class OutputConfiguration(GraderConfiguration):
    """Options for task and result schema."""

    task_details: Dict[str, Any] = field(default_factory=dict)
    result_details: Dict[str, Any] = field(default_factory=dict)

    def add_required_task_detail(self, name: str, default: Any = Nothing):
        """Add a required task detail."""

        self.task_details[name] = default

    def add_required_result_detail(self, name: str, default: Any = Nothing):
        """Add a required task detail."""

        self.result_details[name] = default

    def check_task(self, task: Task):
        """Confirm details are fulfilled."""

        for detail_name, detail_default in self.task_details.items():
            if detail_name not in task.details:
                if detail_default is Nothing:
                    log.error(f"missing required detail {detail_name} in task {task.name}")
                    raise GraderException()
                task.details[detail_name] = detail_default

    def check_result(self, result: Result):
        """Confirm details are fulfilled."""

        for detail_name, detail_default in self.task_details.items():
            if detail_name not in result.details:
                if detail_default is Nothing:
                    log.error(f"missing required detail {detail_name} from result of task {result.task.name}")
                    raise GraderException()
                result.details[detail_name] = detail_default



import functools

from typing import List, Iterable, AnyStr, Union, Sized, Callable, Container, TypeVar, Sequence
from pathlib import Path

from . import CorrectnessResult
from ....library.process import Runtime, Interactive, InteractiveStreamTimeoutExpired, Interaction

__all__ = (
    "as_lines",
    "lines_match",
    "lines_match_test",
    "test_runtime_succeeded",
    "make_stdout_runtime_test",
    "make_in_out_test",
    "make_exit_runtime_test",
    "wrap_runtime_test",
    "make_exit_zero",
    "write_then_read")


def as_lines(string: AnyStr) -> List[AnyStr]:
    """Strip and split by newline."""

    return string.strip().split("\n" if isinstance(string, str) else b"\n")


AnyStrSequence = Union[Iterable[AnyStr], Sized]


def lines_match(a: AnyStrSequence, b: AnyStrSequence) -> bool:
    """Check ordered equality.

    Returns a boolean indicating correctness and a list of errors
    encountered while checking.
    """

    if hasattr(a, "__len__") and hasattr(a, "__len__"):
        if len(a) != len(b):
            return False

    for x, y in zip(a, b):
        if x != y:
            return False

    return True


def lines_match_unordered(a: AnyStrSequence, b: AnyStrSequence) -> bool:
    """Check unordered equality."""

    return lines_match(sorted(a), sorted(b))


def lines_match_test(
        a: Iterable[bytes],
        b: Iterable[bytes],
        ordered: bool = True,
        **details: dict) -> CorrectnessResult:
    """Wrap lines match with correctness result."""

    match = lines_match if ordered else lines_match_unordered
    if match(a, b):
        return CorrectnessResult(passed=True, **details)
    return CorrectnessResult(
        passed=False,
        received=b"\n".join(a).decode() + "\n",
        expected=b"\n".join(b).decode() + "\n",
        **details)


BytesTransform = Callable[[bytes], bytes]
RuntimeTest = Callable[[Runtime], CorrectnessResult]
T = TypeVar("T")


def identity(x: T) -> T:
    return x


def test_runtime_succeeded(runtime: Runtime) -> CorrectnessResult:
    """See if the runtime raised exceptions or returned status code."""

    if runtime.raised_exception:
        return CorrectnessResult(passed=False, runtime=runtime.dump(), error=runtime.exception.description)
    elif runtime.timed_out:
        return CorrectnessResult(passed=False, runtime=runtime.dump(), error="timed out")
    elif runtime.code != 0:
        return CorrectnessResult(passed=False, runtime=runtime.dump(), error="expected status code of zero")
    return CorrectnessResult(passed=True)


def make_in_out_test(
        executable_name: str,
        test_in_path: Path,
        test_out_path: Path,
        *test_out_paths: Path,
        timeout: float = 1) -> Callable[[dict], CorrectnessResult]:
    """Create an input-output test case.

    Returns a test case that feeds the executable the in path on
    the command line and compares its output to the provided
    options.

    Note that by doing it this way, we read the files on grader
    instantiation instead of every time we want to run the test.
    """

    test_out_lines_lists = []
    for out_path in (test_out_path,) + test_out_paths:
        test_out_lines_lists.append(out_path.read_bytes().strip().split(b"\n"))
    compare_stdout = make_stdout_runtime_test(test_out_lines_lists=test_out_lines_lists)

    def test_in_out(resources: dict) -> CorrectnessResult:
        """Actually test the executable."""

        executable = resources[executable_name]
        runtime = executable.execute(str(test_in_path), timeout=timeout)

        # Test failed
        runtime_succeeded = test_runtime_succeeded(runtime)
        if not runtime_succeeded.passed:
            return runtime_succeeded

        return compare_stdout(runtime)

    return test_in_out


def make_stdout_runtime_test(
        *,
        out_transform: BytesTransform = identity,
        out_line_transform: BytesTransform = identity,
        test_out: bytes = None,
        test_out_lines: List[bytes] = None,
        test_out_lines_lists: List[List[bytes]] = None) -> RuntimeTest:
    """Build a test for matching stdout output.

    The list of lines from the stdout of the runtime is compared
    against each list of lines in test_out_line_lists. Note that this
    method first checks whether any error was raised during runtime.
    """

    if len(tuple(filter(lambda x: x is not None, (test_out, test_out_lines, test_out_lines_lists)))) != 1:
        raise ValueError("Exactly one of test_out, test_out_lines, or test_out_lines_lists is required")

    if test_out is not None:
        def compare_stdout(runtime: Runtime):
            """Direct comparision of output."""

            out = out_transform(runtime.stdout)
            passed = out == test_out
            details = {} if passed else dict(
                expected=[test_out.decode(errors="replace")],
                received=out.decode(errors="replace"))
            return CorrectnessResult(passed=passed, runtime=runtime.dump(), **details)
        return compare_stdout

    else:
        if test_out_lines is not None:
            test_out_lines_lists = [test_out_lines]
        expected_out_lines = []
        for expected_out_lines_list in test_out_lines_lists:
            expected_out_lines.append(b"\n".join(expected_out_lines_list).decode(errors="replace"))

        def compare_stdout(runtime: Runtime):
            """Compare by line."""

            out_lines = tuple(map(out_line_transform, out_transform(runtime.stdout).split(b"\n")))
            passed = any(lines_match(out_lines, lines) for lines in test_out_lines_lists)
            details = {} if passed else dict(
                expected=expected_out_lines,
                received=b"\n".join(out_lines))
            return CorrectnessResult(passed=passed, runtime=runtime.dump(), **details)
        return compare_stdout


def make_exit_runtime_test(
        *,
        expected_code: int = None,
        expected_codes: Container[int] = None) -> RuntimeTest:
    """Make a generic exit code test."""

    if expected_code is None and expected_codes is None:
        raise ValueError("Runtime exit test requires either expected status or statuses!")
    if expected_code is not None:
        expected_codes = (expected_code,)

    def test(runtime: Runtime) -> CorrectnessResult:
        return CorrectnessResult(passed=runtime.code in expected_codes, runtime=runtime.dump())

    return test


def wrap_runtime_test(
        *,
        executable_name: str,
        args: Iterable[str] = (),
        runtime_test: RuntimeTest,
        stdin: bytes = None,
        timeout: float = None,
        cwd: Path = None):
    """Make a simple stdout test."""

    def test(resources: dict) -> CorrectnessResult:
        executable = resources[executable_name]
        runtime = executable.execute(*args, stdin=stdin, timeout=timeout, cwd=cwd)

        # Check fail
        runtime_succeeded = test_runtime_succeeded(runtime)
        if not runtime_succeeded.passed:
            return runtime_succeeded

        return runtime_test(runtime)

    return test


exit_zero = functools.partial(wrap_runtime_test, runtime_test=make_exit_runtime_test(expected_code=0))


def make_exit_zero(executable_name: str, args: Iterable[str] = (), stdin: bytes = None, timeout: float = None):
    """Make a test that checks if the program returns zero."""

    return exit_zero(executable_name=executable_name, args=args, stdin=stdin, timeout=timeout)


def write_then_read(
        interactive: Interactive,
        stdin: Sequence[bytes],
        read_condition: Callable[[bytes], bool] = None,
        read_condition_timeout: float = None) -> Interaction:
    """Pass in a bunch of lines, get the last output, compare."""

    if not interactive.poll():
        raise CorrectnessResult(passed=False, error="program crashed")

    with interactive.recording() as interaction:
        for command in stdin:
            interactive.stdin.write(command)
            try:
                interactive.stdout.read(
                    block=True,
                    condition=read_condition,
                    timeout=read_condition_timeout)
            except InteractiveStreamTimeoutExpired:
                raise CorrectnessResult(passed=False, error="timed out", runtime=interaction.dump())

    if interaction.stdout is None:
        raise CorrectnessResult(passed=False, error="did not receive output", runtime=interaction.dump())

    return interaction



import abc
import inspect
from typing import Dict, TypeVar, Generic, Any, Collection, Type
from dataclasses import dataclass, field

from ..log import log


@dataclass(init=False)
class Result(Exception, abc.ABC):
    """The result of a test."""

    complete: bool
    passed: bool
    details: dict

    kind: str = field(init=False)
    task: "Task" = field(init=False)

    def __init__(self, complete: bool, passed: bool, details: dict):
        self.complete = complete
        self.passed = passed
        self.details = details

    def __str__(self):
        return "passed" if self.complete and self.passed else "failed"

    def dump(self) -> dict:
        return dict(
            complete=self.complete,
            passed=self.passed,
            task=self.task.name,
            details=self.details)

    @classmethod
    def incomplete(cls):
        return cls(complete=False, passed=False)


TResult = TypeVar("TResult", bound=Result)


class Runnable(Generic[TResult]):
    def __call__(self, **kwargs) -> TResult:
        ...


def inject(resources: dict, runnable: Runnable[TResult]) -> TResult:
    """Build injection map for method."""

    dependencies = {}
    for name, parameter in inspect.signature(runnable).parameters.items():
        dependency = resources.get(name, parameter.default)
        if dependency == parameter.empty:
            raise ValueError(f"could not satisfy dependency {name}")
        dependencies[name] = dependency
    return runnable(**dependencies)


@dataclass
class Task(Generic[TResult]):
    """Superclass for check, build, run."""

    name: str
    description: str
    stage: str
    kind: str
    dependencies: Collection[str]
    runnable: Runnable[Result]
    details: dict
    result_type: Type[Result]

    def __str__(self):
        return self.name

    def __hash__(self):
        return id(self)

    def run(self, resources: Dict[str, Any]) -> TResult:
        """Do the dependency injection for the runnable."""

        try:
            result = inject(resources, self.runnable)
        except ValueError as error:
            raise ValueError(f"caught in {self.name}: {error}")

        if result is None:
            log.debug(f"task {self.name} did not return a result")
            result = self.result_type()
        return result

    def dump(self):
        """Return the task as JSON serializable."""

        return dict(
            name=self.name,
            description=self.description,
            stage=self.stage,
            kind=self.kind,
            dependencies=self.dependencies,
            details=self.details)



from dataclasses import dataclass

from ...task import Result


@dataclass
class CleanupResult(Result):
    """Deletion of files, deallocation, etc."""

    kind = "cleanup"

    def __init__(self, passed: bool = True, complete: bool = True, **details):
        super().__init__(complete=complete, passed=passed, **details)
